<?php

namespace App\Http\Controllers;

use App\Http\Traits\CanManageFile;
use App\Http\Traits\Response\HasApiResponse;
use Exception;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Database\Eloquent\Model;
use Illuminate\Http\JsonResponse;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Log;
use Illuminate\Support\Str;

abstract class Controller
{
    use CanManageFile, HasApiResponse;

    protected array $relations;

    protected function paginateOrGet(Builder|Model $query) {
        $page = request()->query('page');
        $perPage = request()->query('perPage', 15); // Default perPage to 15 if not provided

        return $query->when($page,
            fn(Builder $q) => $q->paginate($perPage),
            fn(Builder $q) => $q->get()
        );
    }

    protected function getRelations(): array {
        // Get the 'relations' query parameter from the request
        $relationsToLoad = request()->query('relations');

        // If the 'relations' parameter is '*' (all relations), return all predefined relations
        if ($relationsToLoad === '*') {
            return $this->relations;
        }

        // If specific relations are provided, split them into an array and map to camelCase, handling 'factory'
        if ($relationsToLoad) {
            return array_map(
                fn($relation) => Str::camel($relation), explode(',', $relationsToLoad)
            );
        }

        // Return an empty array if no relations are specified
        return [];
    }

    protected function transactional(callable $callback): Model|JsonResponse|null {
        DB::beginTransaction();
        try {
            // Execute the callback
            $result = $callback();

            DB::commit(); // commit the transaction
            return $result; // return the result of the callback
        } catch (Exception $e) {
            DB::rollBack(); // rollback the transaction

            // log the error
            Log::error($e->getMessage());

            return $this->respondError(
                "Action failed: ".$e->getMessage(), $e
            );
        }
    }
}
