<?php

namespace App\Http\Controllers;

use App\Http\Requests\DriverRequest;
use App\Http\Requests\TripRequest;
use App\Http\Requests\VehicleRequest;
use App\Http\Resources\TripResource;
use App\Models\Driver;
use App\Models\Trip;
use App\Models\Vehicle;
use Illuminate\Http\JsonResponse;

class TripController extends Controller
{
    protected array $relations = [
        "transporter", "dispatcher", "vehicle", "driver", "receiver", "ePod",
    ];

    /**
     * Display a listing of the trips.
     */
    public function index(): JsonResponse {
        $trips = $this->paginateOrGet(
            Trip::with($this->getRelations())->filter()->latest());

        return $this->respondWithResourceCollection(
            TripResource::collection($trips)
        );
    }

    /**
     * Store a newly created trip in storage.
     */
    public function store(
        TripRequest $tripRequest,
        VehicleRequest $vehicleRequest,
        DriverRequest $driverRequest
    ): JsonResponse {
        $trip = $this->transactional(function () use ($tripRequest, $vehicleRequest, $driverRequest) {
            $vehicle = Vehicle::updateOrCreate(
                ["vehicle_registration_no" => $vehicleRequest->validated("registration_no")],
                $vehicleRequest->validated()
            );

            $driver = Driver::updateOrCreate(
                ["dl_no" => $driverRequest->validated("dl_no")],
                $driverRequest->validated()
            );

            return Trip::create(
                $tripRequest->validated() + [
                    "vehicle_id" => $vehicle->id,
                    "driver_id" => $driver->id
                ]
            );
        });

        // TODO: send intimation link with trip status mail to the receiver;
        // TODO: based on eta, send trip details mail to the receiver;

        return $this->respondCreated(
            new TripResource($trip),
            "Trip created successfully!"
        );
    }

    /**
     * Display the specified trip.
     */
    public function show(Trip $trip): JsonResponse {
        $trip->loadMissing($this->getRelations());

        return $this->respondWithResource(new TripResource($trip));
    }

    /**
     * Update the specified trip in storage.
     */
    public function update(
        TripRequest $tripRequest,
        VehicleRequest $vehicleRequest,
        DriverRequest $driverRequest,
        Trip $trip
    ): JsonResponse {
        $this->transactional(function () use ($tripRequest, $vehicleRequest, $driverRequest, $trip) {
            $trip->vehicle()->update($vehicleRequest->validated());
            $trip->driver()->update($driverRequest->validated());
            $trip->update($tripRequest->validated());
        });

        return $this->respondUpdated(
            new TripResource($trip),
            "Trip updated successfully!"
        );
    }

    /**
     * Remove the specified trip from storage.
     */
    public function destroy(Trip $trip): JsonResponse {
        $trip->delete();

        return $this->respondSuccess("Trip deleted successfully!");
    }
}
