<?php

namespace App\Models;

use Abbasudo\Purity\Traits\Filterable;
use Illuminate\Database\Eloquent\Factories\HasFactory;
use Illuminate\Database\Eloquent\Model;
use Illuminate\Database\Eloquent\Relations\BelongsTo;
use Illuminate\Database\Eloquent\Relations\HasOne;

class Trip extends Model
{
    use HasFactory, Filterable;

    public $filterFields = [
        'transporter_id',
        'dispatcher_id',
        'vehicle_id',
        'driver_id',
        'receiver_id',

        'is_chassis_no_verified',
        'is_engine_no_verified',
        'is_make_verified',
        'is_model_verified',
        'is_carrying_capacity_verified',
        'is_vehicle_type_verified',
        'is_aadhaar_verified',
        'is_dl_verified',
        'is_dob_verified',
        'is_address_verified',
        'is_mobile_no_verified',
        'dob_remark',
        'address_remark',
        'trip_code',
        'consignor_name',
        'consignee_name',
        'fast_tag_id',
        'dispatch_location',
        'dispatch_pincode',
        'destination_location',
        'destination_pincode',
        'latitude',
        'longitude',
        'eta',
        'distance',
        'is_e_pod_created',
        'status',
    ];

    protected $fillable = [
        'transporter_id',
        'dispatcher_id',
        'vehicle_id',
        'driver_id',
        'receiver_id',

        'is_chassis_no_verified',
        'is_engine_no_verified',
        'is_make_verified',
        'is_model_verified',
        'is_carrying_capacity_verified',
        'is_vehicle_type_verified',
        'is_aadhaar_verified',
        'is_dl_verified',
        'is_dob_verified',
        'is_address_verified',
        'is_mobile_no_verified',
        'dob_remark',
        'address_remark',
        'consignor_name',
        'consignee_name',
        'fast_tag_id',
        'dispatch_location',
        'dispatch_pincode',
        'destination_location',
        'destination_pincode',
        'latitude',
        'longitude',
        'goods_description',
        'eta',
        'distance',
        'is_e_pod_created',
        'status',
    ];

    protected static function boot(): void {
        parent::boot();

        // Before creating a trip, set the trip_code
        static::creating(function ($trip) {
            $trip->trip_code = self::generateTripCode();
        });
    }

    /**
     * Generate a new trip_code with the format ILT11111, ILT11112, ...
     */
    public static function generateTripCode(): string {
        // Get the last trip
        $lastTrip = self::orderBy('id', 'desc')->first();

        // Default starting number (ILT11111)
        $prefix = 'ILT';
        $startNumber = 11111;

        if (!$lastTrip) {
            // No trip yet, return the first one
            return $prefix.$startNumber;
        }

        // Get the numeric part from the last trip_code and increment it
        $lastNumber = (int) substr($lastTrip->trip_code, 3); // Remove the prefix
        $newNumber = $lastNumber + 1;

        // Return the new trip_code in the format ILT11111, ILT11112, etc.
        return $prefix.$newNumber;
    }

    public function transporter(): BelongsTo {
        return $this->belongsTo(Transporter::class);
    }

    public function dispatcher(): BelongsTo {
        return $this->belongsTo(User::class, 'dispatcher_id');
    }

    public function vehicle(): BelongsTo {
        return $this->belongsTo(Vehicle::class);
    }

    // Automatically generate the trip_code

    public function driver(): BelongsTo {
        return $this->belongsTo(Driver::class);
    }

    public function receiver(): BelongsTo {
        return $this->belongsTo(User::class, 'receiver_id');
    }

    public function ePod(): HasOne {
        return $this->hasOne(EPod::class, 'trip_id');
    }

    protected function casts(): array {
        return [
            'is_chassis_no_verified' => 'boolean',
            'is_engine_no_verified' => 'boolean',
            'is_make_verified' => 'boolean',
            'is_model_verified' => 'boolean',
            'is_carrying_capacity_verified' => 'boolean',
            'is_vehicle_type_verified' => 'boolean',
            'is_aadhaar_verified' => 'boolean',
            'is_dl_verified' => 'boolean',
            'is_dob_verified' => 'boolean',
            'is_address_verified' => 'boolean',
            'is_mobile_no_verified' => 'boolean',
            'is_e_pod_created' => 'boolean',
        ];
    }
}
